/*-
 * Copyright 2017 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* for strdup() */
#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <gcrypt.h>
#include <string.h>

#include "vcl.h"
#include "cache/cache.h"
#include "vrt.h"
#include "vas.h"
#include "vdef.h"

#include "vcc_if.h"
#include "vmod_gcrypt.h"
#include "padding.h"

#define str(s) xstr(s)
#define xstr(s) #s

#define MIN_GCRYPT_VERSION "1.6.3"

#define ERR(ctx, msg) \
        errmsg((ctx), "vmod gcrypt error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod gcrypt error: " fmt, __VA_ARGS__)

#define VERRNOMEM(ctx, fmt, ...) \
        VERR((ctx), fmt ", out of space", __VA_ARGS__)

struct vmod_gcrypt_symmetric {
	unsigned		magic;
#define VMOD_GCRYPT_SYMMETRIC_MAGIC 0x82c7ffe2
	gcry_cipher_hd_t	hd;
	char			*vcl_name;
	size_t			blocklen;
	enum padding		padding;
	int			mode;
};

static const char *gcrypt_version = NULL;
static int secmem_enabled = 1;

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		VSB_vprintf(ctx->msg, fmt, args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
	else if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		/* Should this ever happen in vcl_fini() ... */
		VSL(SLT_VCL_Error, 0, fmt, args);
	va_end(args);
}

/* Event function */

static void __match_proto__(gcry_handler_log_t)
gcrypt_logger(void *priv, int level, const char *fmt, va_list args)
{
	enum VSL_tag_e tag = SLT_Debug;
	const char *lvl = "INFO";

	(void) priv;
	switch(level) {
	case GCRY_LOG_CONT:
		lvl = "continuation";
		break;
	case GCRY_LOG_INFO:
		lvl = "INFO";
		break;
	case GCRY_LOG_WARN:
		lvl = "WARN";
		break;
	case GCRY_LOG_ERROR:
		lvl = "ERROR";
		tag = SLT_Error;
		break;
	case GCRY_LOG_FATAL:
		lvl = "FATAL";
		tag = SLT_Error;
		break;
	case GCRY_LOG_BUG:
		lvl = "BUG";
		tag = SLT_Error;
		break;
	case GCRY_LOG_DEBUG:
		lvl = "DEBUG";
		break;
	default:
		VSL(SLT_Error, 0, "Unknown libgcrypt log level %d", level);
	}
	VSL(tag, 0, "libgcrypt log message follows (%s):", lvl);
	VSLv(tag, 0, fmt, args);
}

static void  __match_proto__(gcry_handler_error_t)
gcrypt_fatal(void *priv, int err, const char *text)
{
	(void) priv;
	if (text == NULL)
		text = gpg_strerror(err);
	VSL(SLT_Error, 0, "libgcrypt fatal error, code=%d (%s), panic follows",
	    err, text);
	AZ(text);
}

int __match_proto__(vmod_event_f)
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	(void) ctx;
	(void) priv;

	if (e == VCL_EVENT_LOAD) {
		if (gcrypt_version != NULL)
			return 0;
		gcry_set_log_handler(gcrypt_logger, NULL);
		gcry_set_fatalerror_handler(gcrypt_fatal, NULL);
		gcrypt_version = gcry_check_version(MIN_GCRYPT_VERSION);
		if (gcrypt_version == NULL) {
			VSL(SLT_VCL_Error, 0,
			    "libgcrypt initialization failed");
			return 1;
		}
	}
	return 0;
}

/* Function init */

VCL_VOID
vmod_init(VRT_CTX, VCL_ENUM cmd, VCL_BYTES n)
{
	gcry_error_t err = GPG_ERR_NO_ERROR;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(cmd);
	if (ctx->method != VCL_MET_INIT) {
		ERR(ctx, "gcrypt.init() is only legal in vcl_init");
		return;
	}
	if (gcry_control(GCRYCTL_INITIALIZATION_FINISHED_P)) {
		VSL(SLT_Debug, 0, "libgcrypt initialization already finished");
		return;
	}

	if (strcmp(cmd, "FINISH") == 0) {
		if ((err = gcry_control(GCRYCTL_INITIALIZATION_FINISHED))
		    != GPG_ERR_NO_ERROR)
			VERR(ctx, "Cannot finish initialization in "
			     "gcrypt.init(): %s/%s", gcry_strsource(err),
			     gcry_strerror(err));
		return;
	}
	if (strcmp(cmd, "INIT_SECMEM") == 0) {
		assert(n >= 0);
		if ((err = gcry_control(GCRYCTL_INIT_SECMEM, n))
		    != GPG_ERR_NO_ERROR)
			VERR(ctx, "Cannot initialize secure memory to %d bytes "
			     "in gcrypt.init(): %s/%s", n, gcry_strsource(err),
			     gcry_strerror(err));
		secmem_enabled = n;
		return;
	}
	if (strcmp(cmd, "DISABLE_SECMEM") == 0) {
		if ((err = gcry_control(GCRYCTL_DISABLE_SECMEM))
		    != GPG_ERR_NO_ERROR)
			VERR(ctx, "Cannot disable secure memory in "
			     "gcrypt.init(): %s/%s", gcry_strsource(err),
			     gcry_strerror(err));
		secmem_enabled = 0;
		return;
	}
	WRONG("Illegal cmd enum");
}

/* Object symmetric */

VCL_VOID
vmod_symmetric__init(VRT_CTX, struct vmod_gcrypt_symmetric **symmetricp,
		     const char *vcl_name, VCL_ENUM ciphers, VCL_ENUM modes,
		     VCL_ENUM paddings, VCL_BLOB key, VCL_BOOL secure,
		     VCL_BOOL cbc_cts)
{
	struct vmod_gcrypt_symmetric *symmetric;
	gcry_cipher_hd_t hd;
	int algo = GCRY_CIPHER_NONE, mode = GCRY_CIPHER_MODE_NONE;
	unsigned int flags = 0;
	enum padding padding = _MAX_PADDING;
	gcry_error_t err = GPG_ERR_NO_ERROR;
	size_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(symmetricp);
	AZ(*symmetricp);
	AN(vcl_name);
	AN(ciphers);
	AN(modes);
	AN(paddings);
	if (!gcry_control(GCRYCTL_INITIALIZATION_FINISHED_P)) {
		VERR(ctx,
		     "libgcrypt initialization not finished in %s constructor",
		     vcl_name);
		return;
	}
	if (key == NULL || key->priv == NULL) {
		VERR(ctx, "key is NULL in %s constructor", vcl_name);
		return;
	}
	assert(key->len >= 0);
	if (secure && !secmem_enabled) {
		VERR(ctx, "secure memory not enabled in %s constructor",
		     vcl_name);
		return;
	}

#define CIPHER(e, m) if (strcmp(ciphers, str(e)) == 0) algo = m;
#include "ciphers.h"
#undef CIPHER
	if (algo == GCRY_CIPHER_NONE)
		WRONG("Illegal cipher enum");
	if ((len = gcry_cipher_get_algo_keylen(algo)) == 0) {
		VERR(ctx, "Cannot determine key length for %s cipher in %s "
		     "constructor", ciphers, vcl_name);
		return;
	}
	if ((unsigned)key->len > len) {
		VERR(ctx, "Key length %d is longer than the maximum supported "
		     "length %zu for %s cipher in %s constructor", key->len,
		     len, ciphers, vcl_name);
		return;
	}
	if ((len = gcry_cipher_get_algo_blklen(algo)) == 0) {
		VERR(ctx, "Cannot determine block length for %s cipher in %s "
		     "constructor", ciphers, vcl_name);
		return;
	}

#define MODE(e, m, p, i, c) if (strcmp(modes, str(e)) == 0) mode = m;
#include "modes.h"
#undef MODE
	if (mode == GCRY_CIPHER_MODE_NONE)
		WRONG("Illegal mode enum");

#define PADDING(p) if (strcmp(paddings, str(p)) == 0) padding = p;
#include "paddings.h"
#undef PADDING
	if (padding == _MAX_PADDING)
		WRONG("Illegal padding enum");
	if (!need_padding[mode] || (mode == GCRY_CIPHER_MODE_CBC && cbc_cts))
		padding = NONE;

	if (secure)
		flags |= GCRY_CIPHER_SECURE;
	if (cbc_cts)
		flags |= GCRY_CIPHER_CBC_CTS;

	if ((err = gcry_cipher_open(&hd, algo, mode, flags))
	    != GPG_ERR_NO_ERROR) {
		VERR(ctx, "Cannot open cipher in %s constructor: %s/%s",
		     vcl_name, gcry_strsource(err), gcry_strerror(err));
		return;
	}
	if ((err = gcry_cipher_setkey(hd, key->priv, key->len))
	    != GPG_ERR_NO_ERROR) {
		VERR(ctx, "Cannot set key in %s constructor: %s/%s",
		     vcl_name, gcry_strsource(err), gcry_strerror(err));
		return;
	}

	ALLOC_OBJ(symmetric, VMOD_GCRYPT_SYMMETRIC_MAGIC);
	AN(symmetric);
	*symmetricp = symmetric;
	memcpy(&symmetric->hd, &hd, sizeof(hd));
	symmetric->vcl_name = strdup(vcl_name);
	AN(symmetric->vcl_name);
	symmetric->blocklen = len;
	symmetric->padding = padding;
	symmetric->mode = mode;
}

VCL_VOID
vmod_symmetric__fini(struct vmod_gcrypt_symmetric **symmetricp)
{
	struct vmod_gcrypt_symmetric *symmetric;

	if (symmetricp == NULL)
		return;
	symmetric = *symmetricp;
	if (symmetric == NULL)
		return;
	CHECK_OBJ(symmetric, VMOD_GCRYPT_SYMMETRIC_MAGIC);
	if (symmetric->vcl_name != NULL)
		free(symmetric->vcl_name);
	gcry_cipher_close(symmetric->hd);
	FREE_OBJ(symmetric);
}

VCL_BLOB
vmod_symmetric_encrypt(VRT_CTX, struct vmod_gcrypt_symmetric *symmetric,
		       VCL_BLOB plainblob, VCL_BLOB iv, VCL_BLOB ctr)
{
	size_t len;
	uintptr_t snap;
	void *plaintext;
	struct vmod_priv *ciphertext;
	gcry_error_t err = GPG_ERR_NO_ERROR;
	gcry_cipher_hd_t hd;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(symmetric, VMOD_GCRYPT_SYMMETRIC_MAGIC);
	if (plainblob == NULL || plainblob->priv == NULL) {
		VERR(ctx, "Plaintext BLOB is NULL in %s.encrypt()",
		     symmetric->vcl_name);
		return NULL;
	}

	snap = WS_Snapshot(ctx->ws);
	if ((ciphertext = WS_Alloc(ctx->ws, sizeof(*ciphertext))) == NULL) {
		VERRNOMEM(ctx, "Allocating BLOB result in %s.encrypt()",
			  symmetric->vcl_name);
		return NULL;
	}
	if (symmetric->padding != NONE) {
		plaintext = (padf[symmetric->padding])(ctx->ws, plainblob->priv,
						       plainblob->len,
						       symmetric->blocklen,
						       &len);
		if (plaintext == NULL) {
			VERRNOMEM(ctx, "Allocating padded plaintext in "
				  "%s.encrypt()", symmetric->vcl_name);
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
	}
	else {
		plaintext = plainblob->priv;
		len = plainblob->len;
	}
	if ((ciphertext->priv = WS_Alloc(ctx->ws, len)) == NULL) {
		VERRNOMEM(ctx, "Allocating ciphertext result in %s.encrypt",
			  symmetric->vcl_name);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}

	memcpy(&hd, &symmetric->hd, sizeof(hd));
	if (need_iv[symmetric->mode]) {
		if (iv == NULL) {
			VERR(ctx, "Required initialization vector is NULL in "
			     "%s.encrypt()", symmetric->vcl_name);
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
		assert(iv->len >= 0);
		/* A NULL iv of length 0 is permitted. */
		if (iv->priv == NULL)
			assert(iv->len == 0);
		if ((err = gcry_cipher_setiv(hd, iv->priv, iv->len))
		    != GPG_ERR_NO_ERROR) {
			VERR(ctx, "Cannot set initialization vector in "
			     "%s.encrypt(): %s/%s", symmetric->vcl_name,
			     gcry_strsource(err), gcry_strerror(err));
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
	}
	if (need_ctr[symmetric->mode]) {
		if (ctr == NULL || ctr->priv == NULL) {
			VERR(ctx, "Required counter vector is NULL in "
			     "%s.encrypt()", symmetric->vcl_name);
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
		assert(ctr->len >= 0);
		if ((err = gcry_cipher_setctr(hd, ctr->priv, ctr->len))
		    != GPG_ERR_NO_ERROR) {
			VERR(ctx, "Cannot set counter vector in %s.encrypt(): "
			     "%s/%s", symmetric->vcl_name, gcry_strsource(err),
			     gcry_strerror(err));
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
	}
	if ((err = gcry_cipher_encrypt(hd, ciphertext->priv, len, plaintext,
				       len))
	    != GPG_ERR_NO_ERROR) {
		VERR(ctx, "in %s.encrypt(): %s/%s", symmetric->vcl_name,
		     gcry_strsource(err), gcry_strerror(err));
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	ciphertext->len = len;
	ciphertext->free = NULL;
	return ciphertext;
}

VCL_BLOB
vmod_symmetric_decrypt(VRT_CTX, struct vmod_gcrypt_symmetric *symmetric,
		       VCL_BLOB ciphertext, VCL_BLOB iv, VCL_BLOB ctr)
{
	uintptr_t snap;
	struct vmod_priv *plaintext;
	gcry_error_t err = GPG_ERR_NO_ERROR;
	gcry_cipher_hd_t hd;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(symmetric, VMOD_GCRYPT_SYMMETRIC_MAGIC);
	if (ciphertext == NULL || ciphertext->priv == NULL) {
		VERR(ctx, "Ciphertext BLOB is NULL in %s.decrypt()",
		     symmetric->vcl_name);
		return NULL;
	}
	assert(ciphertext->len >= 0);

	snap = WS_Snapshot(ctx->ws);
	if ((plaintext = WS_Alloc(ctx->ws, sizeof(*plaintext))) == NULL) {
		VERRNOMEM(ctx, "Allocating BLOB result in %s.decrypt()",
			  symmetric->vcl_name);
		return NULL;
	}

	plaintext->priv = WS_Front(ctx->ws);
	if (WS_Reserve(ctx->ws, 0) < (unsigned) ciphertext->len) {
		VERRNOMEM(ctx, "Allocating plaintext result in %s.decrypt()",
			  symmetric->vcl_name);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}

	memcpy(&hd, &symmetric->hd, sizeof(hd));
	if (need_iv[symmetric->mode]) {
		if (iv == NULL) {
			VERR(ctx, "Required initialization vector is NULL in "
			     "%s.decrypt()", symmetric->vcl_name);
			WS_Release(ctx->ws, 0);
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
		assert(iv->len >= 0);
		/* A NULL iv of length 0 is permitted. */
		if (iv->priv == NULL)
			assert(iv->len == 0);
		if ((err = gcry_cipher_setiv(hd, iv->priv, iv->len))
		    != GPG_ERR_NO_ERROR) {
			VERR(ctx, "Cannot set initialization vector in "
			     "%s.decrypt(): %s/%s", symmetric->vcl_name,
			     gcry_strsource(err), gcry_strerror(err));
			WS_Release(ctx->ws, 0);
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
	}
	if (need_ctr[symmetric->mode]) {
		if (ctr == NULL || ctr->priv == NULL) {
			VERR(ctx, "Required counter vector is NULL in "
			     "%s.decrypt()", symmetric->vcl_name);
			WS_Release(ctx->ws, 0);
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
		assert(ctr->len >= 0);
		if ((err = gcry_cipher_setctr(hd, ctr->priv, ctr->len))
		    != GPG_ERR_NO_ERROR) {
			VERR(ctx, "Cannot set counter vector in %s.decrypt(): "
			     "%s/%s", symmetric->vcl_name, gcry_strsource(err),
			     gcry_strerror(err));
			WS_Release(ctx->ws, 0);
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
	}
	if ((err = gcry_cipher_decrypt(hd, plaintext->priv, ciphertext->len,
				       ciphertext->priv, ciphertext->len))
	    != GPG_ERR_NO_ERROR) {
		VERR(ctx, "in %s.decrypt(): %s/%s", symmetric->vcl_name,
		     gcry_strsource(err), gcry_strerror(err));
		WS_Release(ctx->ws, 0);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	if (symmetric->padding == NONE)
		plaintext->len = ciphertext->len;
	else {
		assert(plaintext->len % symmetric->blocklen == 0);
		plaintext->len =
			(unpadlenf[symmetric->padding])(plaintext->priv,
							ciphertext->len,
							symmetric->blocklen);
	}
	WS_Release(ctx->ws, plaintext->len);
	plaintext->free = NULL;
	return plaintext;
}

VCL_STRING
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}

VCL_STRING
vmod_gcrypt_version(VRT_CTX __attribute__((unused)))
{
	AN(gcrypt_version);
	return gcrypt_version;
}
